/* a primitive real-time scheduler 
copyright Victor Yodaiken, Michael Barabanov
History: Barabanov
         modified Oleg Subbotin, Ramesh Nalluri, April 1998
         modified Victor Yodaiken, August 1998
	      took rtl_sched.c and made a pure periodic version.
*/


/* This is the interrupt latency + scheduler latency,
   The user needs to determine this and hardcode it accordingly */



#define ADJUSTMENT  20

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/cons.h>

#include <asm/system.h>
#include <asm/rt_time.h>
#include <asm/rt_irq.h>

static int linux_irq_state;
#define		rtl_make_unrtactive() SFIF = linux_irq_state
#define 	rtl_make_rtactive() {linux_irq_state = SFIF; SFIF = 0;}


#include <linux/kernel.h>
#include <rtl_sched.h>

#ifndef RTL_FPU
#define rtl_fpu_switch(a,b)  ;
#define rtl_fpu_task_init(a,b)  { printk("RTL error: tried to get the \
                                          non fp scheduling module to \
		                          create a fp real time task!\n");\
	                                  kfree(st); return -ENOMEM;}
#define rtl_fpu_init()  ;
#define rtl_fpu_task_delete(x)  ;
#else
static void rtl_fpu_task_init(RT_TASK *, RT_TASK *);
static void rtl_fpu_switch(RT_TASK *, RT_TASK *);
static void rtl_fpu_init(void);
static void rtl_fpu_task_delete(RT_TASK *);
#endif

void debug_resched(RT_TASK * ,RTIME );

#include <asm/segment.h>

RT_TASK *rt_tasks;
RT_TASK *rtl_current;
RT_TASK rtl_linux_task;
#define rtl_is_linux_task(a) (a == &rtl_linux_task)

int rt_startup(void (*fn)(int), int data)
{
	r_sti();
	(*fn)(data);
	rt_task_delete(rtl_current);
	/* should never reach this line */
	return 0;
}
	

int rtl_task_init(RT_TASK *task, void (*fn)(int), int data, int stack_size, 
			int priority,int fp) 
{
	int *st;
	long flags;

	if (task->magic == RT_TASK_MAGIC) {
		return -EINVAL;
	}

	st = (int *) kmalloc(stack_size, GFP_KERNEL);	
	if (!st) {
		return -ENOMEM;
	}

	task->priority = priority; 
	task->magic = RT_TASK_MAGIC; 
	task->state = RT_TASK_DORMANT; 
	task->stack_bottom = st;
	task->stack = st + stack_size / sizeof(int);
        if(fp){
		rtl_fpu_task_init(task,rtl_current);
		}
	else task->uses_fp = 0;
	*--(task->stack) = data;
	*--(task->stack) = (int) fn;
	*--(task->stack) = 0;	/* dummy return address from rt_startup --- Thanks, Victor! */
	*--(task->stack) = (int) rt_startup;

	r_save_flags(flags);
	r_cli();
	task->next = rt_tasks;
	rt_tasks = task;
	r_restore_flags(flags);
	return 0;
}

static unsigned int period=0;
static RT_TIME *task_list =0;
static unsigned int  task_list_size =0;
/* called with hard interrupts disabled */
void rtl_catch_period_interrupt(void){
	RT_TASK *new_t = 0; /* new task to run, if any */
        new_t = task_list[period++];
	if(period >= task_list_size)period = 0;
	if(new_t && new_t->state != RT_TASK_ACTIVE){
		   push_interrupted_task();
		   rtl_make_rtactive();
		   new_t->state = RT_TASK_ACTIVE;
		   rtl_fpu_switch(newt,rtl_current);
		   rtl_switch(new_t);
	}
}

/* called when the current task volunteers to wait */
void rtl_reschedule(void)
{
	RTIME now;
	RT_TASK *t;
	unsigned long flags;
	unsigned int  this_period

	r_save_flags(flags);
	r_cli();

	rtl_current->state = RT_TASK_DELAYED;
	if(!(new_t = pop_interrupted_task())){
		conpr("RTL sched failure: no task to resched\n ");
		newt = &rtl_linux_task;
	} 
	if (rtl_is_linux_task(new_t)){
		rtl_make_unrtactive();
	}
	else rtl_make_rtactive();
	
	rtl_fpu_switch(newt,rtl_current);
	rt_switch_to(newt);
	r_restore_flags(flags);
}



int rt_task_delete(RT_TASK *task)
{
	RT_TASK *t;
	int found = 0, count = 0;
	long flags;

	r_save_flags(flags);
	r_cli();

	if (task->magic != RT_TASK_MAGIC) {
		r_restore_flags(flags);
		return -EINVAL;
	}
	
	if (task != rt_tasks) {
		for (t = rt_tasks; t; t = t->next) {
			if (t->next == task) {
				t->next = task->next;
				found = 1;
				break;
			}
		}
		if (!found) {
			r_restore_flags(flags);
			return -EINVAL;
		}
	} else {
		rt_tasks = task->next;
	}
	task->magic = 0;

	
	/* 
	 *If the current task is the task that is being freed, dont free it
	 * but wait for the scheduler to run and switch it for you 
	 */

	if ( rtl_current == task ){
		rt_set_timer(rt_get_time());
		r_restore_flags(flags);
		/* 
		 * wait for the scheduler to switch the current task and then free
		 */
		while (rtl_current == task){
			if (count++ >= 1000000) {
				conpr("RT_SCHED:PANIC: The task with priority ");
				conprn(task->priority);
				conpr(" has not been switched yet\n");
				return -EINVAL;
			}
		}
		/*
		 * Now happily free it
		 */

		rtl_fpu_task_delete(task);
		kfree(task->stack_bottom);
	}
	else {
		kfree(task->stack_bottom);
		rt_set_timer(rt_get_time());
		r_restore_flags(flags);
	}

	return 0;
}


int rt_task_suspend(RT_TASK *task)
{
	task->state = RT_TASK_DORMANT;
	rt_schedule();
	return 0;
}


int rt_task_wakeup(RT_TASK *task)
{
	task->state = RT_TASK_READY;
	rt_schedule();
	return 0;
}


int rt_task_wait(void)
{
	long flags;
	r_save_flags(flags);
	r_cli();
	rtl_current->state = RT_TASK_DELAYED;
	rtl_current->resume_time += rtl_current->period;
	rt_schedule();
	r_restore_flags(flags);
	return 0;
}

int rt_task_make_periodic(RT_TASK *task, RTIME start_time, RTIME period)
{
	long flags;
	if (task->magic != RT_TASK_MAGIC) {
		return -EINVAL;
	}
	r_save_flags(flags);
	r_cli();
	task->resume_time = start_time;
	task->period = period;
	task->state = RT_TASK_DELAYED;
	rt_set_timer(rt_get_time());
	r_restore_flags(flags);
	return 0;
}

int init_module(void)
{
	rt_tasks = &rtl_linux_task;
	rtl_current = &rtl_linux_task;
	rtl_linux_task.priority = RT_LOWEST_PRIORITY;
	rtl_linux_task.next = 0;
	rtl_linux_task.state = RT_TASK_READY;
	rt_request_timer(&rt_schedule);
	rtl_fpu_init();
	return 0;
}


void cleanup_module(void)
{
	RT_TASK *t;
	RT_TASK *next_t;
	rt_free_timer();
	for (t = rt_tasks; t; t = next_t) {
		next_t = t->next;
		if (t != &rtl_linux_task) {
			rt_task_delete(t);
		}
	}
}



void debug_resched(RT_TASK * nextt,RTIME now) {
		static int debugl=100;
		static int debugt = 0;
		if(nextt->resume_time -now < debugl)
			debugl = nextt->resume_time -now;

		if((debugt > 2000) && (debugl < 25))
		{
			debugt = 0;
			conprn(debugl);
			if(rtl_is_linux_task(rtl_current)){
				conpr("L\n");
			}
			else conpr("R\n");
			debugl = 100;
		}
		else debugt++;
}
		
void static arrange_preempt(RT_TASK *newt){
	RT_TASK *nextt = 0; /* task that will preempt, if any */
	for (t = rt_tasks; t; t = t->next)
		if( (t->state == RT_TASK_DELAYED) &&
			(t->priority < newt->priority) &&
			(!nextt ||(t->resume_time < nextt->resume_time)))
				{
					nextt = t;
				}
	if (nextt)
	{
		rt_set_timer(nextt->resume_time -10 );
		debug_resched(nextt,now);
	}
	else
	{ rt_no_timer(); }
}
